# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utilties to transfer data between Covalent and compute backends
"""

import asyncio
from concurrent.futures import ThreadPoolExecutor
from typing import Dict

from covalent._shared_files import logger
from covalent._shared_files.schemas.asset import AssetUpdate

from ..._dal.result import get_result_object as get_result_object
from .utils import run_in_executor

app_log = logger.app_log
am_pool = ThreadPoolExecutor()


# Consumed by Runner
async def upload_asset_for_nodes(dispatch_id: str, key: str, dest_uris: dict):
    """Typical keys: "output", "deps", "call_before", "call_after", "function"""

    result_object = get_result_object(dispatch_id, bare=True)
    tg = result_object.lattice.transport_graph
    loop = asyncio.get_running_loop()

    futs = []
    for node_id, dest_uri in dest_uris.items():
        if dest_uri:
            node = tg.get_node(node_id)
            asset = node.get_asset(key, session=None)
            futs.append(loop.run_in_executor(am_pool, asset.upload, dest_uri))

    await asyncio.gather(*futs)


async def download_assets_for_node(
    dispatch_id: str, node_id: int, asset_updates: Dict[str, AssetUpdate]
):
    # Keys for src_uris: "output", "stdout", "stderr"

    result_object = get_result_object(dispatch_id, bare=True)
    tg = result_object.lattice.transport_graph
    node = tg.get_node(node_id)
    loop = asyncio.get_running_loop()

    futs = []
    db_updates = {}

    # Mapping from asset key to (non-empty) remote uri
    assets_to_download = {}

    # Prepare asset metadata update; prune empty fields
    for key in asset_updates:
        update = {}
        asset = asset_updates[key].dict()
        if asset["remote_uri"]:
            assets_to_download[key] = asset["remote_uri"]
        # Prune empty fields
        for attr, val in asset.items():
            if val is not None:
                update[attr] = val
        if update:
            db_updates[key] = update

    # Update metadata using the designated DB worker thread
    await run_in_executor(node.update_assets, db_updates)

    for key, remote_uri in assets_to_download.items():
        asset = node.get_asset(key, session=None)
        # Download assets concurrently.
        futs.append(loop.run_in_executor(am_pool, asset.download, remote_uri))
    await asyncio.gather(*futs)



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from covalent.quantum.qserver.database import Database

from .logger import app_log


def get_qelectron_db_path(dispatch_id: str, task_id: int):
    """
    Return the path to the Qelectron database for a given dispatch_id and task_id.

    WARNING: SHOULD ONLY BE USED FROM THE SAME MACHINE
    AS WHERE THE USER'S TASK FUNCTION IS BEING RUN.
    """

    database = Database()

    db_path = database.get_db_path(dispatch_id=dispatch_id, node_id=task_id)

    if db_path.exists():
        app_log.debug(f"Found qelectron DB for task {task_id}")
        return db_path
    else:
        app_log.debug(f"Qelectron database not found for task {task_id}")
        return None



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FastAPI models for /api/v1/resultv2 endpoints"""

from typing import Optional

from pydantic import BaseModel

EDGE_METADATA_KEYS = {
    "edge_name",
    "param_type",
    "arg_index",
}


class EdgeMetadata(BaseModel):
    edge_name: str
    param_type: Optional[str] = None
    arg_index: Optional[int] = None


class EdgeSchema(BaseModel):
    source: int
    target: int
    metadata: EdgeMetadata



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Covalent deploy CLI group."""


import subprocess
from pathlib import Path
from typing import Dict, Tuple

import boto3
import click
from rich.console import Console
from rich.table import Table

from covalent.cloud_resource_manager.core import CloudResourceManager
from covalent.executor import _executor_manager

RESOURCE_ALREADY_EXISTS = "Resources already deployed"
RESOURCE_ALREADY_DESTROYED = "Resources already destroyed"
COMPLETED = "Completed"


def get_crm_object(executor_name: str, options: Dict = None) -> CloudResourceManager:
    """
    Get the CloudResourceManager object.

    Returns:
        CloudResourceManager object.

    """
    executor_module_path = Path(
        __import__(_executor_manager.executor_plugins_map[executor_name].__module__).__path__[0]
    )
    return CloudResourceManager(executor_name, executor_module_path, options)


def get_print_callback(
    console: Console, console_status: Console.status, prepend_msg: str, verbose: bool
):
    """Get print callback method.

    Args:
        console: Rich console object.
        console_status: Console status object.
        prepend_msg: Message to prepend to the output.
        verbose: Whether to print the output inline or not.

    Returns:
        Callback method.

    """
    if verbose:
        return console.print

    def inline_print_callback(msg):
        console_status.update(f"{prepend_msg} {msg}")

    return inline_print_callback


def get_settings_table(crm: CloudResourceManager) -> Table:
    """Get resource provisioning settings table.

    Args:
        crm: CloudResourceManager object.

    Returns:
        Table with resource provisioning settings.

    """
    table = Table(title="Settings")
    table.add_column("Argument", justify="left")
    table.add_column("Value", justify="left")
    for argument in crm.plugin_settings:
        table.add_row(argument, str(crm.plugin_settings[argument]["value"]))
    return table


def get_up_help_table(crm: CloudResourceManager) -> Table:
    """Get resource provisioning help table.

    Args:
        crm: CloudResourceManager object.

    Returns:
        Table with resource provisioning help.

    """
    table = Table()
    table.add_column("Argument", justify="center")
    table.add_column("Required", justify="center")
    table.add_column("Default", justify="center")
    table.add_column("Current value", justify="center")
    for argument in crm.plugin_settings:
        table.add_row(
            argument,
            crm.plugin_settings[argument]["required"],
            str(crm.plugin_settings[argument]["default"]),
            str(crm.plugin_settings[argument]["value"]),
        )
    return table


@click.group(invoke_without_command=True)
def deploy():
    """
    Covalent deploy group with options to:

    1. Spin resources up via `covalent deploy up <executor_name> <options>`.

    2. Tear resources down via `covalent deploy down <executor_name> <options>`.

    3. Show status of resources via `covalent deploy status <executor_name>`.

    4. Show status of all resources via `covalent deploy status`.

    """
    pass


@deploy.command(context_settings={"ignore_unknown_options": True})
@click.argument("executor_name", nargs=1)
@click.argument("vars", nargs=-1)
@click.option(
    "--help", "-h", is_flag=True, help="Get info on default and current values for resources."
)
@click.option("--dry-run", "-dr", is_flag=True, help="Get info on current parameter settings.")
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    help="Show the full Terraform output when provisioning resources.",
)
def up(executor_name: str, vars: Dict, help: bool, dry_run: bool, verbose: bool) -> None:
    """Spin up resources corresponding to executor.

    Args:
        executor_name: Short name of executor to spin up.
        options: Options to pass to the Cloud Resource Manager when provisioning the resources.

    Returns:
        None

    Examples:
        $ covalent deploy up awsbatch --region=us-east-1 --instance-type=t2.micro
        $ covalent deploy up ecs
        $ covalent deploy up ecs --help
        $ covalent deploy up awslambda --verbose --region=us-east-1 --instance-type=t2.micro

    """
    cmd_options = {key[2:]: value for key, value in (var.split("=") for var in vars)}
    if msg := validate_args(cmd_options):
        click.echo(msg)
        return
    crm = get_crm_object(executor_name, cmd_options)
    if help:
        click.echo(Console().print(get_up_help_table(crm)))
        return

    console = Console(record=True)
    prepend_msg = "[bold green] Provisioning resources..."

    with console.status(prepend_msg) as status:
        try:
            crm.up(
                dry_run=dry_run,
                print_callback=get_print_callback(
                    console=console,
                    console_status=status,
                    prepend_msg=prepend_msg,
                    verbose=verbose,
                ),
            )
        except subprocess.CalledProcessError as e:
            click.echo(f"Unable to provision resources due to the following error:\n\n{e}")
            return

    click.echo(Console().print(get_settings_table(crm)))
    exists_msg_with_verbose = "Apply complete! Resources: 0 added, 0 changed, 0 destroyed"
    exists_msg_without_verbose = "found no differences, so no changes are needed"
    export_data = console.export_text()
    if exists_msg_with_verbose in export_data or exists_msg_without_verbose in export_data:
        click.echo(RESOURCE_ALREADY_EXISTS)
    else:
        click.echo(COMPLETED)


@deploy.command()
@click.argument("executor_name", nargs=1)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    help="Show the full Terraform output when spinning down resources.",
)
def down(executor_name: str, verbose: bool) -> None:
    """Teardown resources corresponding to executor.

    Args:
        executor_name: Short name of executor to spin up.

    Returns:
        None

    Examples:
        $ covalent deploy down awsbatch
        $ covalent deploy down ecs --verbose

    """
    crm = get_crm_object(executor_name)

    console = Console(record=True)
    prepend_msg = "[bold green] Destroying resources..."
    with console.status(prepend_msg) as status:
        try:
            crm.down(
                print_callback=get_print_callback(
                    console=console,
                    console_status=status,
                    prepend_msg=prepend_msg,
                    verbose=verbose,
                )
            )
        except subprocess.CalledProcessError as e:
            click.echo(f"Unable to destroy resources due to the following error:\n\n{e}")
            return
    destroyed_msg = "Destroy complete! Resources: 0 destroyed."
    export_data = console.export_text()
    if destroyed_msg in export_data:
        click.echo(RESOURCE_ALREADY_DESTROYED)
    else:
        click.echo(COMPLETED)


# TODO - Color code status.
@deploy.command()
@click.argument("executor_names", nargs=-1, required=False)
def status(executor_names: Tuple[str]) -> None:
    """Show executor resource provision status.

    Args:
        executor_names: Short name(s) of executor to show status for.

    Returns:
        None

    Examples:
        $ covalent deploy status awsbatch
        $ covalent deploy status awsbatch ecs
        $ covalent deploy status

    """
    description = {
        "up": "Provisioned Resources.",
        "down": "No infrastructure provisioned.",
        "*up": "Warning: Provisioning error, retry 'up'.",
        "*down": "Warning: Teardown error, retry 'down'.",
    }

    if not executor_names:
        executor_names = [
            name
            for name in _executor_manager.executor_plugins_map.keys()
            if name not in ["dask", "local", "remote_executor"]
        ]
        click.echo(f"Executors: {', '.join(executor_names)}")

    table = Table()
    table.add_column("Executor", justify="center")
    table.add_column("Status", justify="center")
    table.add_column("Description", justify="center")

    invalid_executor_names = []
    for executor_name in executor_names:
        try:
            crm = get_crm_object(executor_name)
            status = crm.status()
            table.add_row(executor_name, status, description[status])
        except KeyError:
            invalid_executor_names.append(executor_name)

    click.echo(Console().print(table))

    if invalid_executor_names:
        click.echo(
            click.style(
                f"Warning: {', '.join(invalid_executor_names)} are not valid executors.",
                fg="yellow",
            )
        )


def validate_args(args: dict):
    message = None
    if len(args) == 0:
        return message
    if "region" in args and args["region"] != "":
        if not validate_region(args["region"]):
            return f"Unable to find the provided region: {args['region']}"


def validate_region(region_name: str):
    ec2_client = boto3.client("ec2")
    response = ec2_client.describe_regions()
    exists = region_name in [item["RegionName"] for item in response["Regions"]]
    return exists



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""General utils for Covalent."""

import importlib
import inspect
import shutil
import socket
from datetime import timedelta
from typing import Any, Callable, Dict, List, Tuple

import cloudpickle
from pennylane._device import Device

from . import logger
from .config import get_config
from .pickling import _qml_mods_pickle

app_log = logger.app_log
log_stack_info = logger.log_stack_info

DEFAULT_UI_ADDRESS = get_config("user_interface.address")
DEFAULT_UI_PORT = get_config("user_interface.port")


_IMPORT_PATH_SEPARATOR = ":"


def get_ui_url(path):
    baseUrl = f"http://{DEFAULT_UI_ADDRESS}:{DEFAULT_UI_PORT}"
    return f"{baseUrl}{path}"


def get_random_available_port() -> int:
    """
    Return a random port that is available on the machine
    """
    sock = socket.socket()
    sock.bind(("127.0.0.1", 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


def get_timedelta(time_limit: str) -> timedelta:
    """
    Get timedelta from a compatible time limit string passed to the lattice/electron decorator.

    Args:
        time_limit: The time limit string.
    Returns:
        timedelta: The `datetime.timedelta` object.
    """

    days, hours, minutes, seconds = time_limit.split("-")[0], *time_limit.split("-")[1].split(":")
    return timedelta(
        days=int(days),
        hours=int(hours),
        minutes=int(minutes),
        seconds=int(seconds),
    )


def reformat(t: int) -> str:
    """
    Reformat an integer to a readable time-like string. For example, if t = 1, return "01".

     Args:
         t: The integer to reformat.

     Returns:
         ref_string: The reformatted string.
    """

    return f"0{t}" if len(str(t)) == 1 else str(t)


def get_time(time_delta: timedelta) -> str:
    """
    Get a compatible time string from a timedelta object.

    Args:
        time_delta: The timedelta object.
    Returns:
        time_string: The compatible reformatted time string.
    """

    days = reformat(time_delta.days)
    hours = reformat(time_delta.seconds // 3600)
    minutes = reformat((time_delta.seconds // 60) % 60)
    seconds = reformat(time_delta.seconds % 60)
    return f"{days}-{hours}:{minutes}:{seconds}"


def filter_null_metadata(meta_dict: dict) -> Dict:
    """Filter out metadata that is None or empty."""
    return {k: v for k, v in meta_dict.items() if v}


def get_serialized_function_str(function):
    """
    Generates a string representation of a function definition
    including the decorators on it.

    Args:
        function: The function whose definition is to be convert to a string.

    Returns:
        function_str: The string representation of the function definition.
    """

    input_function = function
    # If a Lattice or electron object was passed as the function input, we need the
    # (deserialized) underlying function describing the lattice.
    while hasattr(input_function, "workflow_function"):
        input_function = input_function.workflow_function.get_deserialized()

    try:
        # function_str is the string representation of one function, with decorators, if any.
        function_str = inspect.getsource(input_function)
    except Exception:
        function_str = f"# {function.__name__} was not inspectable"

    return function_str + "\n\n"


def get_imports(func: Callable) -> Tuple[str, List[str]]:
    """
    Given an input workflow function, find the imports that were used, and determine
        which ones are Covalent-related.

    Args:
        func: workflow function.

    Returns:
        A tuple consisting of a string of import statements and a set of names that
            Covalent-related modules have been imported as.
    """

    imports_str = ""
    cova_imports = []
    for i, j in func.__globals__.items():
        if inspect.ismodule(j) or (
            inspect.isfunction(j) and j.__name__ in ["lattice", "electron"]
        ):
            if j.__name__ == i:
                import_line = f"import {j.__name__}\n"
            else:
                import_line = f"import {j.__name__} as {i}\n"

            if j.__name__ in ["covalent", "lattice", "electron"]:
                import_line = f"# {import_line}"
                cova_imports.append(i)

            imports_str += import_line

    return imports_str, cova_imports


def required_params_passed(func: Callable, kwargs: Dict) -> bool:
    """
    DEPRECATED: Check to see that values for all parameters without default values have been passed.

    Args:
        func: Callable function.
        kwargs: Parameter list with passed values.

    Returns:
        status: Whether all the parameters required for the callable function has been passed.
    """

    required_arg_set = set({})
    sig = inspect.signature(func)
    for param in sig.parameters.values():
        if param.default is param.empty:
            required_arg_set.add(str(param))

    return required_arg_set.issubset(set(kwargs.keys()))


def get_named_params(func, args, kwargs):
    ordered_params_dict = inspect.signature(func).parameters

    named_args = {}
    named_kwargs = {}

    for ind, parameter_dict in enumerate(ordered_params_dict.items()):
        param_name, param = parameter_dict

        if param.kind in [param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD]:
            if param_name in kwargs:
                named_kwargs[param_name] = kwargs[param_name]
            elif ind < len(args):
                named_args[param_name] = args[ind]
        elif param.kind == param.VAR_POSITIONAL:
            for i in range(ind, len(args)):
                named_args[f"arg[{i}]"] = args[i]
        elif param.kind == param.VAR_KEYWORD:
            for key, value in kwargs.items():
                if key != param_name:
                    named_kwargs[key] = value
        elif param.kind == param.KEYWORD_ONLY:
            if param_name in kwargs:
                named_kwargs[param_name] = kwargs[param_name]

    if len(args) > len(named_args):
        raise ValueError(
            f"Too many positional arguments given, expected {len(named_args)}, received {len(args)}"
        )

    if len(kwargs) > len(named_kwargs):
        extra_supplied_kwargs = ", ".join(sorted(set(kwargs.keys()) - set(named_kwargs.keys())))
        raise ValueError(f"Unexpected keyword arguments: {extra_supplied_kwargs}")

    return (named_args, named_kwargs)


def format_server_url(hostname: str = None, port: int = None) -> str:
    if hostname is None:
        hostname = get_config("dispatcher.address")
    if port is None:
        port = int(get_config("dispatcher.port"))

    url = hostname
    if not url.startswith("http"):
        url = f"https://{url}" if port == 443 else f"http://{url}"

    # Inject port
    if port not in [80, 443]:
        parts = url.split("/")
        url = "".join(["/".join(parts[:3])] + [f":{port}/"] + ["/".join(parts[3:])])

    return url.strip("/")


# For use by LocalDispatcher and ResultsManager when running Covalent
# server locally
def copy_file_locally(src_uri, dest_uri):
    scheme_prefix = "file://"
    if src_uri.startswith(scheme_prefix):
        src_path = src_uri[len(scheme_prefix) :]
    else:
        raise TypeError(f"{src_uri} is not a valid URI")
        # src_path = src_uri
    if dest_uri.startswith(scheme_prefix):
        dest_path = dest_uri[len(scheme_prefix) :]
    else:
        raise TypeError(f"{dest_uri} is not a valid URI")

    shutil.copyfile(src_path, dest_path)


@_qml_mods_pickle
def cloudpickle_serialize(obj):
    return cloudpickle.dumps(obj)


def cloudpickle_deserialize(obj):
    return cloudpickle.loads(obj)


def select_first_executor(qnode, executors):
    """Selects the first executor to run the qnode"""
    return executors[0]


def get_import_path(obj) -> Tuple[str, str]:
    """
    Determine the import path of an object.
    """
    module = inspect.getmodule(obj)
    if module:
        module_path = module.__name__
        class_name = obj.__name__
        return f"{module_path}{_IMPORT_PATH_SEPARATOR}{class_name}"
    raise RuntimeError(f"Unable to determine import path for {obj}.")


def import_from_path(path: str) -> Any:
    """
    Import a class from a path.
    """
    module_path, class_name = path.split(_IMPORT_PATH_SEPARATOR)
    module = importlib.import_module(module_path)
    return getattr(module, class_name)


def get_original_shots(dev: Device):
    """
    Recreate vector of shots if device has a shot vector.
    """
    if not dev.shot_vector:
        return dev.shots

    shot_sequence = []
    for shots in dev.shot_vector:
        shot_sequence.extend([shots.shots] * shots.copies)
    return type(dev.shot_vector)(shot_sequence)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Settings request and response model"""

import enum
from typing import Union

from pydantic import BaseModel


class GetSettingsResponseModel(BaseModel):
    """Get Settings response model"""

    client: Union[dict, None] = None
    server: Union[dict, None] = None


class UpdateSettingsResponseModel(BaseModel):
    """Update Settings response model"""

    data: str


class Validators(enum.Enum):
    """Validate settings overwrite"""

    DISPATCHER = "dispatcher"
    DASK = "dask"
    WORKFLOW_DATA = "workflow_data"
    USER_INTERFACE = "user_interface"
